package fhg2a.relogo

/**
 * Farmed cereal is a food resource maintained by handling (cultivation by a tribe)
 * rather than automatic growth. Growth is based on tribe activity and wild
 * cereal population. Handling and net acquisition rate are calculated based on
 * the time taken to produce the cereal rather than a 'search' time, as is the
 * case with the basic Food type.
 * 
 */
class FarmedCereal extends Food {

	def capacity = 0.0         // FIXME: units / sq km?
	def farmCost = 6.0         // Cal per minute
	def NAR                    // FIXME: Cal per min?
	def totalTimePerUnit       // minutes
	def isFarmed			   // boolean - defines farming status of this resource

	/**
	 * Set the upper limit (number of units) for this resource
	 * based on the carrying capacity, and initialise this
	 * resource as unfarmed.
	 * @param carryingCapacity Maximum density in units per sq km
	 */
	def init(carryingCapacity) {

		capacity = carryingCapacity

		//farmed cereal growth defined by formula in paper
		def bonusCapacity = (0.25 * (1000.0 - capacity))

		//set the maximum amount of farmed cereal available based on
		//calculated capacity and area (normally set to 10% of other areas)
		maximum = (capacity + bonusCapacity) * area

		setFarm(false)

		/* NET ACQUISITION RATE INITIALISATION */
		/* Sets global variables required for farming calculations */

		//time to produce a kilo of cereal
		//(equivalent of search time)
		def sqKmFarmedPerMinute = 0.02 //FIXME: is this now just ABscape value? was (1.0 / 30.0)
		def productionTime = 60.0 * 365.0 / (sqKmFarmedPerMinute * 100 * (capacity + bonusCapacity)) // minutes

		//total time is equal to production time + harvest time (harvest time here is identical to that of wild cereal handle time)
		totalTimePerUnit = productionTime + handleTime

		def energyReaped = energyPerUnit / totalTimePerUnit
		def productionEnergy = (productionTime / totalTimePerUnit) * farmCost
		def handleEnergy = (handleTime / totalTimePerUnit) * farmCost

		//calculates NAR similar to that of other food but some variables are different for farmed cereal - see class doc string
		NAR = energyReaped - farmCost

		System.out.println("capacity / bonus:" + capacity + " / " + bonusCapacity)
		System.out.println("sq km farmed per minute:" + sqKmFarmedPerMinute)
		System.out.println("production minutes:" + productionTime)
		System.out.println("handle minutes:" + handleTime)
		System.out.println("total minutes:" + totalTimePerUnit)
		System.out.println("base energy:" + energyReaped)
		System.out.println("cost:" + farmCost)
		System.out.println("NAR:" + NAR)
	}

	/**
	 * Set this resource to a farmed or unfarmed state.
	 * @param isFarming Flag to indicate if farming is occurring.
	 */
	def setFarm(isFarming) {

		//if we are farming on already farmed land 
		if (isFarming && isFarmed) {
			//do nothing
			System.out.println("Still farming.")
		}

		//if we are the first farmers, make farmed cereal available
		else if (isFarming && !isFarmed) {
			current = 0.0
			isFarmed = true
			System.out.println("Started farming. " + (maximum - current) + " units available.")
		}

		//otherwise, deactivate farming (and remove all farmed cereal)
		else {
			current = 0.0
			isFarmed = false
		}
	}

	/**
	 * Hide grow method. Farmed cereal does not grow, it must be cultivated
	 * by calling handle().
	 */
	def abstract grow
	
	/**
	 * Deactivate farming. Used to reset farming every epoch. Farming will
	 * only be reactivated if a tribe chooses to farm here.
	 */
	 def deactivate() {
		 setFarm(false)
	 }

	 /**
	  * Cultivate farmed cereal and returns net energy obtained by tribe
	  * in 365 14-hour days. Tribe will never gather more than it needs or
	  * more than is available.
	  * @param energyRequired Maximum energy required by handling attempt.
	  * @param populationSize The number of individuals available to handle food (ignored in farmed cereal handling).
	  * @return Energy obtained from farmed cereal in one year
	  */
	 def handle(energyRequired, populationSize) {

		 //calculate maximum number of units available based on the carrying capacity
		 def availableUnits = Math.max(0.0, maximum - current)
		 
		 //if zero energy is required (usually indicates farming is being ignored)
		 //or the farming capacity has been reached, return 0
		 if (energyRequired == 0.0 || availableUnits == 0.0) {
			 return 0.0
		 }
		 //otherwise, mark this resource as farmed
		 else {
			 setFarm(true)
		 }

		 def requiredUnits = Math.max(0, Math.ceil(energyRequired / NAR))

		 //calculate maximum number of units obtainable based on time taken to obtain them
		 def obtainableUnits = Math.floor((365.0 * 14.0 * 60.0) / totalTimePerUnit)

		 //calculate actual number obtained given food and tribe population limits
		 def obtainedUnits = Math.max(0, Math.min(availableUnits, Math.min(requiredUnits, obtainableUnits)))

		 //add units obtained to running total (units farmed while this resource has been active)
		 current += obtainedUnits

		 System.out.println("Farmed " + obtainedUnits + "/" + requiredUnits + " kilos, leaving " + (maximum - current) + " available.")

		 //return net energy obtained
		 return obtainedUnits * NAR
	 }

	 /**
	  * Get the net acquisition rate for this resource.
	  * @return Net acquisition rate for this resource.
	  */
	 def netAcquisitionRate() {
	 	return NAR
	 }
}
