package fhg2a.relogo

/**
 * Renewable resource for tribes to extract energy from via handling (hunting,
 * gathering or farming).
 *
 */
class Food {

	//FIXME: change catchCost to handleCost

	//parameters of food type
	def energyPerUnit          // Cal per unit
	def handleTime             // minutes per unit
	def growthRate             // units per year
	def area                   // sq km
	def searchRadius = 0.0175  // km 
	def searchSpeed = 0.5 / 60 // km per min
	def catchCost = 6          // Cal per min
	def searchCost = 4         // Cal per min

	//maximum units ever available, and current units available
	def maximum = 0.0          // units (= carrying capacity * area)
	def current = 0.0          // units

	/**
	 * Set the upper limit (number of units) for this food source
	 * based on the carrying capacity, and initialise current
	 * amount to that maximum.
	 * @param carryingCapacity Maximum number of units per sq km
	 * 
	 */
	def init(carryingCapacity) {
		maximum = carryingCapacity * area
		current = maximum
	}

	/**
	 * Grow food source based on current amount, growth rate and
	 * upper limit.
	 * 
	 */
	def grow() {

		/* This implementation is based on Dorscheidt (ABscape) rather
		 than van der Vaart et al */

		//calculate predicted amount of food at end of growth cycle
		def numerator = current * maximum * (Math.pow(Math.E, growthRate))
		def denominator = maximum - (current * (1 - Math.pow(Math.E, growthRate)))
		def predictedAmount

		//if denominator is non-zero, calculate predicted amount and
		//ensure it is within 0 and max bounds
		if (denominator != 0) {
			predictedAmount = Math.floor(numerator / denominator)
			predictedAmount = Math.min(maximum, predictedAmount)
			predictedAmount = Math.max(0, predictedAmount)
		}

		//otherwise, denominator too small - set predicted amount to maximum
		else {
			predictedAmount = maximum
		}

		//set current wild cereal population to calculated amount
		current = predictedAmount
	}

	/**
	 * Decrement food resource and return net energy obtained by tribe
	 * in 365 14-hour days. Tribe will never gather more than it needs or
	 * more than is available.
	 * @param maximumEnergyRequired Maximum energy required by handling attempt.
	 * @param populationSize The number of individuals available to handle food.
	 * @return Energy obtained from this resource in one year
	 * 
	 */
	def handle(maximumEnergyRequired, populationSize) {

		//if food is depleted, no energy can be gathered
		//or, if zero energy is required, no energy need be returned
		if (current == 0.0 || maximumEnergyRequired == 0.0) {
			return 0.0
		}

		def foodDensity = current / area

		//search time formula
		def searchTime = 1 / (searchSpeed * searchRadius * 2 * foodDensity)

		//calculate maximum number of food units required
		//(population will not take more than this)
		//in cases where food is sparse, net energy per unit can become negative,
		//in which case the required units are zero as it is not worth it for
		//the tribe to search
		def costOfObtaining = (searchTime * 4) + (handleTime * 6)
		def netEnergyPerUnit = energyPerUnit - costOfObtaining
		def required = Math.max(0, Math.ceil(maximumEnergyRequired / netEnergyPerUnit))

		//calculate maximum number of food units theoretically obtainable
		//by a tribe of this size in 365 days of 14 hour handles
		def obtainable = Math.floor((365 * 14 * 60 * populationSize) / (searchTime + handleTime))

		//calculate actual number obtained given the limits of available food,
		//available workers and energy requirements
		def obtained = Math.min(Math.min(current, required), obtainable)

		//decrement food by number of units obtained
		current = current - obtained

		//return net energy obtained
		System.out.println("Handled " + obtained + " units. " + current + "/" + maximum + " left.")
		return obtained * netEnergyPerUnit
	}

	/**
	 * Calculate the NAR for this food resource.
	 * @returns Net acquisition rate
	 * 
	 */
	def  netAcquisitionRate(){

		/* This implementation follows the van der Vaart formula. */

		//if food is depleted then formula cannot be calculated - see searchTime below (would mean division by 0)
		if (current == 0.0) {
			return 0.0
		}

		def foodDensity = current / area

		//search time formula
		def searchTime = 1 / (searchSpeed * searchRadius * 2 * foodDensity)
		def totalTime = searchTime + handleTime
		def energyPerMinute = energyPerUnit / totalTime
		def searchTimeRatio = searchTime / totalTime
		def handleTimeRatio = handleTime / totalTime

		//calculates NAR (Cal/minute) according to formula in paper
		return energyPerMinute - (searchTimeRatio * searchCost) - (handleTimeRatio * catchCost)
	}
}
