package fhg2a.relogo

import groovy.lang.MetaClass;

/**
 * Tribe that relies on wild cereal, prey and farmed cereal
 * for sustenance, and has perfect global knowledge. Divides
 * annual foraging time between all food types based on
 * their relative net acquisition rates.
 * 
 */
class IdealFreeDistribution extends Tribe {

	/**
	 * Forage one year's worth of prey, wild cereal and farmed cereal.
	 * Decrease population if enough food is not gathered.
	 * 
	 */
	 def forage() {

		 //get acquisition rates for all food types
		 def preyNAR = patchHere().prey.netAcquisitionRate()
		 def wildCerealNAR = patchHere().wildCereal.netAcquisitionRate()
		 def farmedCerealNAR = patchHere().farmedCereal.netAcquisitionRate()

		 //if farming is less efficient than gathering wild cereal,
		 //don't farm (all farm calculations will return 0)
		 if (farmedCerealNAR <= wildCerealNAR) {
			 farmedCerealNAR = 0.0
		 }

		 //initialise measures of energy
		 def energyRequired = populationSize * 2000.0 * 365.0
		 def energyGained = 0.0

		 //calculate proportions of effort for prey based on energy efficiency
		 //then get energy from prey
		 def totalNAR = preyNAR + wildCerealNAR + farmedCerealNAR
		 def portion = preyNAR / totalNAR
		 energyGained += patchHere().handlePrey(portion * energyRequired, populationSize)
		 energyRequired = Math.max(0.0, energyRequired - energyGained)

		 //recalculate proportions of effort for wild cereal based on energy efficiency
		 //and energy needs based on prey success, then get energy from wild cereal
		 totalNAR = wildCerealNAR + farmedCerealNAR
		 portion =  wildCerealNAR / totalNAR
		 energyGained += patchHere().handleWildCereal(portion * energyRequired, populationSize)
		 energyRequired = Math.max(0.0, energyRequired - energyGained)

		 //if farming is viable
		 if (farmedCerealNAR > 0.0) {

			 //get energy from farmed cereal with energy needs based on prey and wild cereal success
			 energyGained += patchHere().handleFarmedCereal(energyRequired, populationSize)
			 energyRequired = Math.max(0.0, energyRequired - energyGained)
		 }

		 //cull tribe population according to total energy obtained from all sources
		 cull(energyGained)
	 }

	 /**
	  * Move tribe to the best patch on the grid (including patch it is on).
	  * Patches are ranked by prey density then cereal density.
	  * 
	  */
	 def move() {

		 //get the set of patches in the world that have the maximum
		 //PREY density of all patches on the grid
		 def bestPreySet = patches().withMax() {
			 it.prey.current / 300.0
		 }

		 //get subset of this set that have the maximum
		 //WILD CEREAL density
		 def bestWildCerealSet = bestPreySet.withMax() {
			 it.wildCereal.current / 300.0
		 }

		//move to the first patch in the subset
		def patchNow = patchHere()
		moveTo(bestWildCerealSet.get(0))
		def patchAfter = patchHere()

		//count a migration (movement between habitats) if it occurred
		if(!patchNow.habitat.equalsIgnoreCase(patchAfter.habitat)) {
			migrations++
		}
	}
}
