package fhg2a.relogo

import static repast.simphony.relogo.Utility.*;
import static repast.simphony.relogo.UtilityG.*;
import repast.simphony.relogo.Plural;
import repast.simphony.relogo.Stop;
import repast.simphony.relogo.Utility;
import repast.simphony.relogo.UtilityG;
import repast.simphony.relogo.schedule.Go;
import repast.simphony.relogo.schedule.Setup;
import fhg2a.ReLogoTurtle;

/**
 * Abstract turtle forming the basis for all tribes (agents).
 * All tribes reproduce and cull according to the same rules
 * and these methods are therefore defined here. Foraging
 * and movement algorithms must be implemented by each
 * extending class.
 * 
 */
abstract class Tribe extends ReLogoTurtle {

	// initial population size
	def populationSize = 20 // number of individuals

	// annual statistics
	def deaths = 0          // number of individuals
	def births = 0          // number of individuals
	def migrations = 0      // number of migrations

	/**
	* Execute tribe's yearly cycle of activity in a fixed order:
	* reproduce, decide whether to move to a new patch,
	* forage for food. Also reset annual statistics.
	* @param tribeType Type of daughter tribes to generate when reproducing (if any).
	* 
	*/
	def step(tribeType) {

		// reset annual statistics
		births = 0
		deaths = 0
		migrations = 0
	
		reproduce(tribeType)
		move()
		forage()
	}

	/**
	 * Increase population of tribe. Each individual has a 2% chance of reproducing.
	 * Tribe will split into daughter tribes of size 20 if population size exceeds 39.
	 * @param tribeType Type of daughter tribes to generate (if any).
	 * 
	 */
	def reproduce(tribeType) {

		//no new babies at start of reproduction cycle
		def babies = 0

		//each person in population has 2% chance of producing a new person
		for (person in 1..populationSize) {
			if (random(99)<=1) {
				babies++
				births++
			}
		}

		//add babies to population count
		populationSize += babies

		//if population exceeds max of 39
		if (populationSize >= 40) {

			//calculate number of daughter tribes to split into
			def daughterTribes = (int) Math.ceil((populationSize - 20) / 20)

			//create daughter tribe clones
			hatch(daughterTribes, {it -> reset()},tribeType)

			//adjust population to remainder after daughter tribes have left
			populationSize = populationSize - (daughterTribes * 20)
		}
	}

	/**
	 * Sets a tribe's population back to the default (20 individuals).
	 * 
	 */
	def reset() {
		populationSize = 20.0
	}

	/**
	 * All extending classes must implement a foraging algorithm.
	 * 
	 */
	def abstract forage()
	
	/**
	 * Cull population to maximum size that can be sustained by
	 * energy gathered.
	 * @param availableEnergy Amount of energy available to support the tribe population.
	 * 
	 */
	def cull(availableEnergy) {

		//calculate number of people that can be feed with available energy
		def sustainablePopulation = Math.floor(availableEnergy / (2000.0 * 365.0))

		//trim population size to sustainable number
		if (populationSize > sustainablePopulation){
			deaths += (populationSize - sustainablePopulation)
			populationSize = sustainablePopulation
		}

		//if no individuals left in the tribe, kill the tribe
		if (populationSize == 0)
			die()
	}

	/**
	 * All extending classes must implement an algorithm for deciding whether
	 * or not to move, and to which patch.
	 */
	def abstract move()

	/* STATISTICS COLLECTION METHODS */
	
	/**
	 * Determine the number of births on patch (if patch is lush).
	 * @return Number of births if habitat is lush, or 0 otherwise.
	 * 
	 */
	def birthsOnLush() {
		if (patchHere().habitat.equalsIgnoreCase("Lush")) {
			return births
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the number of births on patch (if patch is middle habitat).
	 * @return Number of births if habitat is middle, or 0 otherwise.
	 * 
	 */
	def birthsOnMiddle(){
		if (patchHere().habitat.equalsIgnoreCase("Middle")) {
			return births
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of births on patch (if patch is desert).
	 * @return Number of births if habitat is desert, or 0 otherwise.
	 * 
	 */
	def birthsOnDesert() {
		// FIXME: "desert" should be "arid"
		if (patchHere().habitat.equalsIgnoreCase("Desert")) {
			return births
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the number of deaths on patch (if patch is lush).
	 * @return Number of deaths if habitat is lush, or 0 otherwise.
	 * 
	 */
	def deathsOnLush() {
		if (patchHere().habitat.equalsIgnoreCase("Lush")) {
			return deaths
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the number of deaths on patch (if patch is middle habitat).
	 * @return Number of deaths if habitat is middle, or 0 otherwise.
	 * 
	 */
	def deathsOnMiddle() {
		if (patchHere().habitat.equalsIgnoreCase("Middle")) {
			return deaths
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of deaths on patch (if patch is desert).
	 * @return Number of deaths if habitat is desert, or 0 otherwise.
	 * 
	 */
	def deathsOnDesert() {
		if (patchHere().habitat.equalsIgnoreCase("Desert")) {
			return deaths
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the number of agents on patch (if patch is lush).
	 * @return Number of agents if habitat is lush, or 0 otherwise.
	 * 
	 */
	def agentOnLush() {
		//FIXME: is this agents (tribes) or individuals?
		if (patchHere().habitat.equalsIgnoreCase("Lush")) {
			return populationSize
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of agents on patch (if patch is middle habitat).
	 * @return Number of agents if habitat is middle, or 0 otherwise.
	 * 
	 */
	def agentOnMiddle() {
		//FIXME: is this agents (tribes) or individuals?
		if (patchHere().habitat.equalsIgnoreCase("Middle")) {
			return populationSize
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of agents on patch (if patch is desert).
	 * @return Number of agents if habitat is desert, or 0 otherwise.
	 * 
	 */
	def agentOnDesert() {
		//FIXME: is this agents (tribes) or individuals?
		if (patchHere().habitat.equalsIgnoreCase("Desert")) {
			return populationSize
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of migrations by this agent (if patch is
	 * lush).
	 * @return Number of agents if habitat is lush, or 0 otherwise.
	 * 
	 */
	def migrationsToLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return migrations
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the number of migrations by this agent (if patch is
	 * middle habitat).
	 * @return Number of agents if habitat is middle, or 0 otherwise.
	 * 
	 */
	def migrationsToMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return migrations
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the number of migrations by this agent (if patch is
	 * desert).
	 * @return Number of agents if habitat is desert, or 0 otherwise.
	 * 
	 */
	def migrationsToDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return migrations
		}
		else {
			return 0
		}
	}

}
