package fhg2a.relogo

import static repast.simphony.relogo.Utility.*;
import static repast.simphony.relogo.UtilityG.*;
import repast.simphony.relogo.Stop;
import repast.simphony.relogo.Utility;
import repast.simphony.relogo.UtilityG;
import repast.simphony.relogo.ast.Diffusible;
import repast.simphony.relogo.schedule.Go;
import repast.simphony.relogo.schedule.Setup;
import fhg2a.ReLogoPatch;

/**
 * A single patch of terrain on the grid with food resources for tribes
 * to handle (hunt, gather or farm).
 *
 */
class UserPatch extends ReLogoPatch {

	def area = 300.0           // sq km
	def habitat = "Undefined"  // lush / middle / desert // FIXME: enum? arid?

	//food sources with energy, handling and growth parameters
	//(Cal per unit, minutes per unit, units per year) #FIXME: confirm unit for rate?
	def prey = Food.newInstance([energyPerUnit: 13800.0, handleTime: 235.0, growthRate: 0.7, area: 300.0])
	def wildCereal = Food.newInstance([energyPerUnit: 3390.0, handleTime: 120.0, growthRate: 1.0, area: 300.0])
	def farmedCereal = FarmedCereal.newInstance([energyPerUnit: 3390.0, handleTime: 120.0, growthRate: 0.0, area: 30.0])

	/**
	 * Set upper limit for each food resources based on carrying capacity
	 * (maximum sustainable density), and initializes number of food units
	 * to maximum for each food type (based on area of).
	 * @param preyPerSqKm Maximum prey density
	 * @param wildCerealPerSqKm Maximum wild cereal density
	 * @param farmedCerealPerSqKm Maximum farmed cereal density
	 */
	def setCarryingCapacity(preyPerSqKm, wildCerealPerSqKm, farmedCerealPerSqKm) {

		//convert density (units per square kilometer) to actual amount on patch (units)
		prey.init(preyPerSqKm)
		wildCereal.init(wildCerealPerSqKm)
		farmedCereal.init(farmedCerealPerSqKm)
	}

	/**
	 * Decrement prey and return net energy obtained by tribe
	 * in 365 14-hour days. Tribe will never gather more than it needs or
	 * more than is available.
	 * @param maximumEnergyRequired Maximum energy required by handling attempt.
	 * @param populationSize The number of individuals available to handle food.
	 * @return Energy obtained from prey in one year
	 * 
	 */
	def handlePrey(maximumEnergyRequired, populationSize) {
		def gatheredEnergy = prey.handle(maximumEnergyRequired, populationSize)
		return gatheredEnergy
	}

	/**
	 * Decrement wild cereal and return net energy obtained by tribe
	 * in 365 14-hour days. Tribe will never gather more than it needs or
	 * more than is available.
	 * @param maximumEnergyRequired Maximum energy required by handling attempt.
	 * @param populationSize The number of individuals available to handle food.
	 * @return Energy obtained from wild cereal in one year
	 * 
	 */
	def handleWildCereal(maximumEnergyRequired, populationSize) {
		def gatheredEnergy = wildCereal.handle(maximumEnergyRequired, populationSize)
		return gatheredEnergy
	}

	/**
	 * Cultivate farmed cereal and return net energy obtained by tribe
	 * in 365 14-hour days. Tribe will never gather more than it needs or
	 * more than is available.
	 * @param energyRequired Maximum energy required by handling attempt.
	 * @param populationSize The number of individuals available to handle food (ignored in farmed cereal handling).
	 * @return Energy obtained from farmed cereal in one year
	 * 
	 */
	def handleFarmedCereal(maximumEnergyRequired, populationSize) {
		def gatheredEnergy = farmedCereal.handle(maximumEnergyRequired, populationSize)
		return gatheredEnergy
	}

	/**
	 * Replenish prey resource and update NAR value.
	 * 
	 */
	def replenishPrey() {
		prey.grow()
	}

	/**
	 * Replenish wild cereal resource and update NAR value.
	 * 
	 */
	def replenishWildCereal() {
		wildCereal.grow()
	}

	/**
	 * Replenish and hide farmed cereal capacity. Farmed cereal
	 * will only become visible again if a tribe farms here.
	 * 
	 */
	def deactivateFarmedCereal() {
		farmedCereal.deactivate()
	}

	/* STATISTICS COLLECTION METHODS */

	/**
	 * Determine the total amount of wild cereal on this patch.
	 * @return Total wild cereal units
	 * 
	 */
	def totalWildCereal() {
		return wildCereal.current
	}

	/**
	 * Determine the total amount of prey on this patch.
	 * @return Total prey units
	 * 
	 */
	def totalPrey() {
		return prey.current
	}
	
	/**
	 * Determine the total amount of farmed cereal on this patch.
	 * @return Total farmed cereal units
	 * 
	 */
	def totalFarmedCereal() {
		return farmedCereal.current
	}

	/**
	 * Determine the total amount of wild cereal on patch (if patch is lush).
	 * @return Total wild cereal units if habitat is lush, or 0 otherwise.
	 * 
	 */
	def wildCerealOnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return wildCereal.current
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the total amount of wild cereal on patch (if patch is middle habitat).
	 * @return Total wild cereal units if habitat is middle, or 0 otherwise.
	 * 
	 */
	def wildCerealOnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return wildCereal.current
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the total amount of wild cereal on patch (if patch is desert).
	 * @return Total wild cereal units if habitat is desert, or 0 otherwise.
	 * 
	 */
	def wildCerealOnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return wildCereal.current
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the total amount of prey on patch (if patch is lush).
	 * @return Total prey units if habitat is lush, or 0 otherwise.
	 * 
	 */
	def preyOnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return prey.current
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the total amount of prey on patch (if patch is middle habitat).
	 * @return Total prey units if habitat is middle, or 0 otherwise.
	 * 
	 */
	def preyOnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return prey.current
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the total amount of prey on patch (if patch is desert).
	 * @return Total prey units if habitat is desert, or 0 otherwise.
	 * 
	 */
	def preyOnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return prey.current
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the total amount of farmed cereal on patch (if patch is lush).
	 * @return Total farmed cereal units if habitat is lush, or 0 otherwise.
	 * 
	 */
	def farmedCerealOnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return farmedCereal.current
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the total amount of farmed cereal on patch (if patch is middle habitat).
	 * @return Total farmed cereal units if habitat is middle, or 0 otherwise.
	 * 
	 */
	def farmedCerealOnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return farmedCereal.current
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the total amount of farmed cereal on patch (if patch is desert).
	 * @return Total farmed cereal units if habitat is desert, or 0 otherwise.
	 * 
	 */
	def farmedCerealOnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return farmedCereal.current
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the wild cereal NAR on this patch.
	 * @return Wild cereal net acquisition rate
	 * 
	 */
	def wildCerealTotalNAR() {
		return wildCereal.netAcquisitionRate()
	}
	
	/**
	 * Determine the prey NAR on this patch.
	 * @return Prey net acquisition rate
	 * 
	 */
	def preyTotalNAR() {
		return prey.netAcquisitionRate()
	}
	
	
	/**
	 * Determine the farmed cereal NAR on this patch.
	 * @return Farmed cereal net acquisition rate
	 * 
	 */
	def farmedCerealTotalNAR() {
		if (farmedCereal.netAcquisitionRate()!=null) {
			return farmedCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the wild cereal NAR on this patch (if patch is lush).
	 * @return Wild cereal net acquisition rate if habitat is lush, or 0 otherwise.
	 * 
	 */
	def wildCerealNAROnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return wildCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the wild cereal NAR on this patch (if patch is middle habitat).
	 * @return Wild cereal net acquisition rate if habitat is middle, or 0 otherwise.
	 * 
	 */
	def wildCerealNAROnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return wildCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the wild cereal NAR on this patch (if patch is desert).
	 * @return Wild cereal net acquisition rate if habitat is desert, or 0 otherwise.
	 * 
	 */
	def wildCerealNAROnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return wildCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the prey NAR on this patch (if patch is lush).
	 * @return Prey net acquisition rate if habitat is lush, or 0 otherwise.
	 * 
	 */
	def preyNAROnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return prey.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the prey NAR on this patch (if patch is middle habitat).
	 * @return Prey net acquisition rate if habitat is middle, or 0 otherwise.
	 * 
	 */
	def preyNAROnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return prey.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the prey NAR on this patch (if patch is desert).
	 * @return Prey net acquisition rate if habitat is desert, or 0 otherwise.
	 * 
	 */
	def preyNAROnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return prey.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the farmed cereal NAR on this patch (if patch is lush).
	 * @return Farmed cereal net acquisition rate if habitat is lush, or 0 otherwise.
	 * 
	 */
	def farmedCerealNAROnLush() {
		if (habitat.equalsIgnoreCase("Lush")) {
			return farmedCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Determine the farmed cereal NAR on this patch (if patch is middle habitat).
	 * @return Farmed cereal net acquisition rate if habitat is middle, or 0 otherwise.
	 * 
	 */
	def farmedCerealNAROnMiddle() {
		if (habitat.equalsIgnoreCase("Middle")) {
			return farmedCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}
	
	/**
	 * Determine the farmed cereal NAR on this patch (if patch is desert).
	 * @return Farmed cereal net acquisition rate if habitat is desert, or 0 otherwise.
	 * 
	 */
	def farmedCerealNAROnDesert() {
		if (habitat.equalsIgnoreCase("Desert")) {
			return farmedCereal.netAcquisitionRate()
		}
		else {
			return 0
		}
	}

	/**
	 * Repopulates empty prey and wild cereal resources based on the
	 * number of units on neighboring patches. Farmed cereal cannot
	 * spontaneously repopulate, it must be handled (cultivated) by
	 * a tribe.
	 * 
	 */
	def crosspopulateFood(){
		//FIXME: this contains duplicate code, should be refactored

		//flag to indicate whether patch must be repopulated
		//(lower flag at start)
		def repopulate = false

		//if patch has no more cereal
		if(wildCereal.current == 0){
			//poll neighbors
			ask(neighbors()){
				//test neighbour's viability
				def isViable = (it.wildCereal.current >= 400)
				//attempt a 10% chance
				def isLucky = (random(9) == 0)
				//if viable neighbour and attempt sucessful, raise flag to repopulate
				//or keep flag raised if raised already
				repopulate = repopulate || (isViable && isLucky)
			}

			//if marked for repopulation, set amount of wild cereal to viable level
			if (repopulate){
				wildCereal.current = 400
			}
		}

		//if patch has no more prey
		repopulate = false
		if(prey.current == 0){
			//poll neighbors
			ask(neighbors()){
				//test neighbour's viability
				def isViable = (it.prey.current >= 100)
				//attempt a 10% chance
				def isLucky = (random(9) == 0)
				//if viable neighbour and attempt sucessful, raise flag to repopulate
				//or keep flag raised if raised already
				repopulate = repopulate || (isViable && isLucky)
			}

			//if marked for repopulation, set amount of prey to viable level
			if (repopulate){
				prey.current = 100
			}
		}
	}
}
